"""Цей модуль перетворює вираз з операціями,
заданими в інфіксній формі,
в обернений польский запис.
Обернений польський запис виразу -
це список елементів виразу,
у якому операнди передують операції."""

from exprlib.common import priority


def is_letter(sym):
    """Визначає, чи символ sym є малою латинською літерою."""
    return "a" <= sym <= "z"


def is_digit(sym):
    """Визначає, чи символ sym є цифрою."""
    return "0" <= sym <= "9"


def peek(stack):
    """Видає останній елемент непустого списку stack."""
    ns = len(stack) - 1
    fs = stack[ns]
    return fs


def dijkstra(mode, op, stackop, res):
    """Реалізує алгоритм Дейкстри.
    mode - це тип лексеми:
    u - унарна операція,
    b - бінарна операція,
    ( - дужка,
    v - змінна
    n - число
    e - заключна обробка стеку операцій.
    stackop - це стек операцій,
    res - це результуючий стек елементів виразу."""
    if mode == "v":
        res.append(op)
    elif mode == "n":
        num = int(op)
        res.append(num)
    elif mode == "e":
        ops = stackop.pop()
        res.append(ops)
    else:
        # унарні, бінарні операції або дужки
        if len(stackop) == 0:
            stackop.append(op)
        elif op == "(":
            stackop.append(op)
        elif op == "-" or op == "+":
            # Унарна операція
            stackop.append(op)
        else:
            last = peek(stackop)
            prlast = priority(last)
            priop = priority(op)
            if priop == 0:
                stackop.append(op)
            elif op == ")" and last == "(":
                stackop.pop()
            elif priop > prlast:
                stackop.append(op)
            else:
                while len(stackop) > 0 and priop <= prlast:
                    ops = stackop.pop()
                    res.append(ops)
                    if len(stackop) > 0:
                        last = peek(stackop)
                        prlast = priority(last)
                if last == "(" and op == ")":
                    stackop.pop()
                else:
                    stackop.append(op)


def polish(expr):
    """Здійснюючи лексичний аналіз виразу expr,
    записаного в інфіксній формі,
    готуючи для застосування алгоритму Дейкстри
    для перетворення в обернений польський запис."""
    stackop = []
    res = []
    ident = ""
    num = ""
    op = ""
    for sym in expr:
        if sym == "(":
            dijkstra("(", sym, stackop, res)
        elif sym == ")" or sym == " ":
            if len(ident) > 0:
                # Ідентифікатор
                if ident == "not":
                    # Унарна літерна операція
                    dijkstra("u", ident + " ", stackop, res)
                    ident = ""
                elif ident == "and" or ident == "or":
                    # Бінарна літерна операція
                    dijkstra("b", " " + ident + " ", stackop, res)
                    ident = ""
                else:
                    # Змінна
                    if len(op) > 0 and op != " ":
                        # Унарна операція
                        dijkstra("u", op, stackop, res)
                        # op = ""
                    # elif sym == " ":
                    if sym == " ":
                        op = " "
                    dijkstra("v", ident, stackop, res)
                    ident = ""
            elif len(num) > 0 and op != " ":
                # Число
                if len(op) > 0:
                    # Унарна операція
                    dijkstra("u", op, stackop, res)
                    op = ""
                if sym == " ":
                    op = " "
                dijkstra("n", num, stackop, res)
                num = ""
            elif len(op) > 0:
                # Завершується бінарна операція
                dijkstra("b", op + " ", stackop, res)
                op = ""
            elif sym == " ":
                # Починається бінарна операція
                op = " "
            if sym == ")":
                dijkstra("(", sym, stackop, res)
                op = ""
        elif is_letter(sym):
            ident += sym
        elif is_digit(sym):
            if ident == "":
                num += sym
            else:
                ident += sym
        elif sym == "+" or sym == "-" or sym == "*":
            if op == " ":
                # Бінарна операція
                op = " " + sym
            elif sym == "+" or sym == "-":
                if len(op) > 0:
                    # Декілька унарних операцій підряд
                    dijkstra("u", op, stackop, res)
                op = sym
        elif sym == "=":
            op += sym
        elif sym == "!":
            op += sym
        elif sym == "<":
            op += sym
        elif sym == ">":
            op += sym
    if len(ident) > 0:
        dijkstra("v", ident, stackop, res)
        if len(op) > 0:
            # Унарна операція
            dijkstra("u", op, stackop, res)
    if len(num) > 0:
        dijkstra("n", num, stackop, res)
        if len(op) > 0:
            # Унарна операція
            dijkstra("u", op, stackop, res)
    while len(stackop) > 0:
        dijkstra("e", op, stackop, res)
    return res



"""Побудова дерев виразів"""

from exprlib.polish import polish
from exprlib.common import operation, priority, unop


class Tree():
    """Клас дерев виразів."""
    def __init__(self):
        self.op = None
        self.lt = None
        self.rt = None

    def getop(self):
        return self.op

    def getleft(self):
        return self.lt

    def getright(self):
        return self.rt

    def setop(self, op):
        self.op = op

    def setleft(self, left):
        self.lt = left

    def setright(self, right):
        self.rt = right

    def unary(self):
        # Чи операція, розташована в корені дерева, є унарною?
        return self.lt is None

    def binary(self):
        # Чи операція, розташована в корені дерева, є бінарною?
        return self.lt is not None

    def tostring(self):
        res = treeexpr(self)
        return res


def createtree(op, lt, rt):
    """Утворює дерево з бінарною операцією op і піддеревами lt та rt."""
    etree = Tree()
    etree.setop(op)
    etree.setleft(lt)
    etree.setright(rt)
    return etree


def poltree(polexpr):
    """Перетворює вираз expr, заданий оберненим польським записом, в дерево виразу.
    Служить ще одним конструктором класу Tree."""
    if len(polexpr) == 1:
        return polexpr[0]
    etree = Tree()
    stack = []
    for el in polexpr:
        if operation(el):
            etree = Tree()
            etree.setop(el)
            right = stack.pop()
            #if len(stack) > 0:
            if not unop(el):
                left = stack.pop()
                etree.setleft(left)
            etree.setright(right)
            stack.append(etree)
        else:
            stack.append(el)
    return etree


def exprtree(expr):
    """Перетворює вираз expr, заданий у інфіксній формі, в дерево виразу
     через обернений польський запис. Служить ще одним конструктором класу Tree."""
    polexpr = polish(expr)
    etree = poltree(polexpr)
    return etree


def treeexpr(etree, fop="(", lr="lt"):
    """Перетворює дерево виразу etree у вираз інфіксної форми.
    fop - це операція батьківського виразу.
    lr - її лівий операнд ("lt") або правий ("rt").
    Рекурсивна функція."""
    if (etree is True) or (etree is False):
        return str(etree)
    if type(etree) is str:
        return etree
    elif type(etree) is int:
        return str(etree)
    else:
        op = etree.getop()
        rt = etree.getright()
        if (op == "True") or (op == "False"):
            return op
        if etree.unary():
            # Унарна операція
            subtree = op + treeexpr(rt, op, "rt")
            if priority(fop) > priority(op):
                return "(" + subtree + ")"
            else:
                return subtree
        else:
            # Бінарна операція
            lt = etree.getleft()
            subtree = treeexpr(lt, op, "lt") + op + treeexpr(rt, op, "rt")
            if fop == "-":
                return subtree
            elif priority(fop) > priority(op):
                return "(" + subtree + ")"
            elif fop == " - ":
                if op == " * ":
                    subtree = treeexpr(lt, op, "lt") + op + treeexpr(rt, op, "rt")
                else:
                    if lr == "lt":
                        subtree = treeexpr(lt, op, "lt") + op + treeexpr(rt, op, "rt")
                    else:
                        subtree = "(" + treeexpr(lt, op, "lt") + op + treeexpr(rt, op, "rt") + ")"
                return subtree
            else:
                return subtree


def expr_expr(expr):
    """Перетворює заданий вираз expr в дерево і назад."""
    tree = exprtree(expr)
    rexpr = treeexpr(tree)
    return rexpr




def operation(el):
    """Визначає, чи текст el є операцією."""
    if el == "+" or el == "-":
        # Унарні операції
        return True
    elif el == " + " or el == " - ":
        # Бінарні операції
        return True
    elif el == " * ":
        return True
    return False


def priority(op):
    """Визначає пріоритет операції op."""
    if op == "(":
        return 0
    if op == ")":
        return 1
    if op == " + ":
        return 6
    if op == " - ":
        return 6
    if op == " * ":
        return 7
    if op == "+":
        # Унарний +
        return 8
    if op == "-":
        # Унарний -
        return 8
    return 10


def unop(op):
    """Визначає, чи є op унарною операцією."""
    if op == "-":
        return True
    if op == "+":
        return True
    return False


from setuptools import setup, find_packages

setup(name='exprlib',
      version='0.1a4',
      url='https://github.com/crocodile850/ExprLib',
      # license='MIT',
      author='LuteVasyl',
      author_email='crocodile21@gmail.com',
      packages=find_packages(exclude=['tests']),
      # description='Convert expression from infix form to binary tree and vice versa.',
      # long_description=open('README.md').read(),
      # classifiers=[
      #     "Programming Language :: Python :: 3",
      #     "License :: OSI Approved :: MIT License",
      # ],
      # python_requires='>=3.6'
      )
